/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
import { oob } from "@fluidframework/core-utils/internal";
import { Tree } from "./shared-tree/index.js";
import { TreeArrayNode, } from "./simple-tree/index.js";
// Future improvement TODOs (ideally to be done before promoting these APIs to `@alpha`):
// - Custom fields on Table/Row/Column (props pattern from Nick's demo)
// - Overloads to make Column/Row schema optional when constructing Tables
// - Record-like type parameters / input parameters?
// - Move `@system` types into separate / sub scope?
/**
 * Contains types and factories for creating schema to represent dynamic tabular data.
 * @privateRemarks TODO: document in more detail and add `@example`s.
 * @internal
 */
export var TableSchema;
(function (TableSchema) {
    const tableSchemaFactorySubScope = "table";
    /**
     * Factory for creating new table column schema.
     * @privateRemarks
     * TODO:
     * - Add overloads to make propsSchema optional.
     * - Take field schema rather than node schema for `propsSchema`, in particular to allow making
     * the additional properties optional.
     * @internal
     */
    // eslint-disable-next-line @typescript-eslint/explicit-function-return-type -- Return type is too complex to be reasonable to specify
    function createColumn(inputSchemaFactory, propsSchema) {
        const schemaFactory = inputSchemaFactory.scopedFactory(tableSchemaFactorySubScope);
        /**
         * {@link Column} fields.
         * @remarks Extracted for re-use in returned type signature defined later in this function.
         * The implicit typing is intentional.
         * Note: ideally we would add a satisfies clause here to ensure that this satisfies
         * `Record<string, ImplicitFieldSchema>`, but doing so causes TypeScript to prematurely and incorrectly evaluate the type of `propsSchema`.
         * Likely related to the following issue: https://github.com/microsoft/TypeScript/issues/52394
         */
        const columnFields = {
            id: schemaFactory.identifier,
            props: schemaFactory.required(propsSchema),
        };
        /**
         * A column in a table.
         */
        class Column extends schemaFactory.object("Column", columnFields) {
        }
        // Returning SingletonSchema without a type conversion results in TypeScript generating something like `readonly "__#124291@#brand": unknown;`
        // for the private brand field of TreeNode.
        // This numeric id doesn't seem to be stable over incremental builds, and thus causes diffs in the API extractor reports.
        // This is avoided by doing this type conversion.
        // The conversion is done via assignment instead of `as` to get stronger type safety.
        const ColumnSchemaType = Column;
        return ColumnSchemaType;
    }
    TableSchema.createColumn = createColumn;
    /**
     * Factory for creating new table row schema.
     *
     * @privateRemarks
     * TODO:
     * - Add overloads to make propsSchema optional.
     * - Take field schema rather than node schema for `propsSchema`, in particular to allow making
     * the additional properties optional.
     *
     * @sealed @internal
     */
    // eslint-disable-next-line @typescript-eslint/explicit-function-return-type -- Return type is too complex to be reasonable to specify
    function createRow(inputSchemaFactory, cellSchema, propsSchema) {
        const schemaFactory = inputSchemaFactory.scopedFactory(tableSchemaFactorySubScope);
        /**
         * {@link Row} fields.
         * @remarks Extracted for re-use in returned type signature defined later in this function.
         * The implicit typing is intentional.
         * Note: ideally we would add a satisfies clause here to ensure that this satisfies
         * `Record<string, ImplicitFieldSchema>`, but doing so causes TypeScript to prematurely and incorrectly evaluate the type of `propsSchema`.
         * Likely related to the following issue: https://github.com/microsoft/TypeScript/issues/52394
         */
        const rowFields = {
            id: schemaFactory.identifier,
            cells: schemaFactory.map("Row.cells", cellSchema),
            props: schemaFactory.required(propsSchema),
        };
        /**
         * The Row schema - this is a map of Cells where the key is the column id
         */
        class Row extends schemaFactory.object("Row", rowFields) {
            getCell(columnOrId) {
                const columnId = typeof columnOrId === "string" ? columnOrId : columnOrId.id;
                return this.cells.get(columnId);
            }
            setCell(columnOrId, value) {
                const columnId = typeof columnOrId === "string" ? columnOrId : columnOrId.id;
                this.cells.set(columnId, value);
            }
            removeCell(columnOrId) {
                const columnId = typeof columnOrId === "string" ? columnOrId : columnOrId.id;
                if (!this.cells.has(columnId)) {
                    return;
                }
                this.cells.delete(columnId);
            }
        }
        // Returning SingletonSchema without a type conversion results in TypeScript generating something like `readonly "__#124291@#brand": unknown;`
        // for the private brand field of TreeNode.
        // This numeric id doesn't seem to be stable over incremental builds, and thus causes diffs in the API extractor reports.
        // This is avoided by doing this type conversion.
        // The conversion is done via assignment instead of `as` to get stronger type safety.
        const RowSchemaType = Row;
        return RowSchemaType;
    }
    TableSchema.createRow = createRow;
    /** `createTable` implementation */
    function createTable(inputSchemaFactory, _cellSchema, columnSchema, rowSchema) {
        const column = columnSchema ?? createColumn(inputSchemaFactory, inputSchemaFactory.null);
        return createTableInternal(inputSchemaFactory, _cellSchema, column, rowSchema ??
            createRow(inputSchemaFactory, _cellSchema, inputSchemaFactory.null));
    }
    TableSchema.createTable = createTable;
    /**
     * Factory for creating new table schema.
     * @system @internal
     */
    // eslint-disable-next-line @typescript-eslint/explicit-function-return-type -- Return type is too complex to be reasonable to specify
    function createTableInternal(inputSchemaFactory, _cellSchema, columnSchema, rowSchema) {
        const schemaFactory = inputSchemaFactory.scopedFactory(tableSchemaFactorySubScope);
        /**
         * {@link Table} fields.
         * @remarks Extracted for re-use in returned type signature defined later in this function.
         * The implicit typing is intentional.
         */
        const tableFields = {
            rows: schemaFactory.array("Table.rows", rowSchema),
            columns: schemaFactory.array("Table.columns", columnSchema),
        };
        /**
         * The Table schema
         */
        class Table extends schemaFactory.object("Table", tableFields) {
            getColumn(id) {
                // TypeScript is unable to narrow the types correctly here, hence the casts.
                // See: https://github.com/microsoft/TypeScript/issues/52144
                return this.columns.find((column) => column.id === id);
            }
            getRow(id) {
                // TypeScript is unable to narrow the types correctly here, hence the casts.
                // See: https://github.com/microsoft/TypeScript/issues/52144
                return this.rows.find((_row) => _row.id === id);
            }
            getCell(key) {
                const { columnId, rowId } = key;
                const row = this.getRow(rowId);
                if (row !== undefined) {
                    const column = this.getColumn(columnId);
                    if (column !== undefined) {
                        return row.getCell(column.id);
                    }
                }
                // If the cell does not exist return undefined
                return undefined;
            }
            insertColumn({ column, index, }) {
                if (index === undefined) {
                    // TypeScript is unable to narrow the types correctly here, hence the cast.
                    // See: https://github.com/microsoft/TypeScript/issues/52144
                    // eslint-disable-next-line @typescript-eslint/no-explicit-any
                    this.columns.insertAtEnd(column);
                }
                else {
                    // TypeScript is unable to narrow the types correctly here, hence the cast.
                    // See: https://github.com/microsoft/TypeScript/issues/52144
                    // eslint-disable-next-line @typescript-eslint/no-explicit-any
                    this.columns.insertAt(index, column);
                }
                // Inserting the input node into the tree hydrates it, making it usable as a node.
                return column;
            }
            insertRows({ index, rows, }) {
                if (index === undefined) {
                    // TypeScript is unable to narrow the types correctly here, hence the cast.
                    // See: https://github.com/microsoft/TypeScript/issues/52144
                    // eslint-disable-next-line @typescript-eslint/no-explicit-any
                    this.rows.insertAtEnd(TreeArrayNode.spread(rows));
                }
                else {
                    // TypeScript is unable to narrow the types correctly here, hence the cast.
                    // See: https://github.com/microsoft/TypeScript/issues/52144
                    // eslint-disable-next-line @typescript-eslint/no-explicit-any
                    this.rows.insertAt(index, TreeArrayNode.spread(rows));
                }
                // Inserting the input nodes into the tree hydrates them, making them usable as nodes.
                return rows;
            }
            setCell({ key, cell }) {
                const { columnId, rowId } = key;
                const row = this.getRow(rowId);
                if (row !== undefined) {
                    const column = this.getColumn(columnId);
                    if (column !== undefined) {
                        row.setCell(column.id, cell);
                    }
                }
            }
            removeColumn(column) {
                const index = this.columns.indexOf(column);
                // If the column is not in the table, do nothing
                if (index === -1)
                    return;
                this.columns.removeAt(index);
            }
            removeRows(rows) {
                // If there are no rows to remove, do nothing
                if (rows.length === 0) {
                    return;
                }
                // If there is only one row to remove, remove it
                if (rows.length === 1) {
                    const index = this.rows.indexOf(rows[0] ?? oob());
                    this.rows.removeAt(index);
                    return;
                }
                // If there are multiple rows to remove, remove them in a transaction
                // This is to avoid the performance issues of deleting multiple rows at once
                Tree.runTransaction(this, () => {
                    // Iterate over the rows and remove them
                    for (const row of rows) {
                        const index = this.rows.indexOf(row);
                        this.rows.removeAt(index);
                    }
                });
            }
            removeAllRows() {
                this.rows.removeRange();
            }
            removeCell(key) {
                const { columnId, rowId } = key;
                const row = this.getRow(rowId);
                if (row !== undefined) {
                    const column = this.getColumn(columnId);
                    if (column !== undefined) {
                        row.removeCell(column.id);
                    }
                }
            }
        }
        // Returning SingletonSchema without a type conversion results in TypeScript generating something like `readonly "__#124291@#brand": unknown;`
        // for the private brand field of TreeNode.
        // This numeric id doesn't seem to be stable over incremental builds, and thus causes diffs in the API extractor reports.
        // This is avoided by doing this type conversion.
        // The conversion is done via assignment instead of `as` to get stronger type safety.
        const TableSchemaType = Table;
        // Return the table schema
        return TableSchemaType;
    }
    TableSchema.createTableInternal = createTableInternal;
    // #endregion
})(TableSchema || (TableSchema = {}));
//# sourceMappingURL=tableSchema.js.map